#ifndef INVERT_H_
#define INVERT_H_
#include <complex>
#include <fstream>
#include <iostream>
#include <iomanip>
#include "petscksp.h"
using namespace std;

const double PI = 3.141592635;
void permuteRow(complex<double> [4][4], int, int);
void permuteColumn(complex<double> [4][4], int, int);

#undef __FUNCT__
#define __FUNCT__ "InvertHamiltonian"
PetscErrorCode InvertHamiltonian (Mat hamiltonian, Mat inverseHamiltonian, Mat identityMat, PC pc, KSP ksp, block blockHamiltonian[300][3])
{
	//PETSc Setup
	const PetscInt rows = 3600;
	PetscErrorCode ierr;
	
	//Variable Declarations
	complex<double> fermi1, fermi2, fermif, transmissionFunction;
	complex<double> scaledTransmissionFunction;
	
	//Fill Hamiltonian
	for(int j=0; j<300; j++)
	{
		for(int k=0; k<12; k++)
		{
			for(int l=0; l<12; l++)
			{
				if(blockHamiltonian[j][1].getData(k, l) != complex<double>(0,0))
					ierr = MatSetValue(hamiltonian, k+12*j, l+12*j, blockHamiltonian[j][1].getData(k, l), INSERT_VALUES);CHKERRQ(ierr);
			} //endfor
		} //endfor
	} //endfor

	for(int j=0; j<299; j++)
	{
		for(int k=0; k<12; k++)
		{
			for(int l=0; l<12; l++)
			{
				if(blockHamiltonian[j][2].getData(k, l) != complex<double>(0,0))
					ierr = MatSetValue(hamiltonian, k+12*j, l+12*(j-1), blockHamiltonian[j][2].getData(k, l), INSERT_VALUES);CHKERRQ(ierr);
				if(j == 298)
				{
					if(blockHamiltonian[j][2].getData(0, 1) != complex<double>(0,0))
						ierr = MatSetValue(hamiltonian, rows-12, rows-23, blockHamiltonian[j][2].getData(0, 1), INSERT_VALUES);CHKERRQ(ierr);
					if(blockHamiltonian[j][2].getData(3, 2) != complex<double>(0,0))
						ierr = MatSetValue(hamiltonian, rows-9, rows-22, blockHamiltonian[j][2].getData(3, 2), INSERT_VALUES);CHKERRQ(ierr);
					if(blockHamiltonian[j][2].getData(4, 5) != complex<double>(0,0))
						ierr = MatSetValue(hamiltonian, rows-8, rows-19, blockHamiltonian[j][2].getData(4, 5), INSERT_VALUES);CHKERRQ(ierr);
					if(blockHamiltonian[j][2].getData(7, 6) != complex<double>(0,0))
						ierr = MatSetValue(hamiltonian, rows-5, rows-18, blockHamiltonian[j][2].getData(7, 6), INSERT_VALUES);CHKERRQ(ierr);
					if(blockHamiltonian[j][2].getData(8, 9) != complex<double>(0,0))
						ierr = MatSetValue(hamiltonian, rows-4, rows-15, blockHamiltonian[j][2].getData(8, 9), INSERT_VALUES);CHKERRQ(ierr);
					if(blockHamiltonian[j][2].getData(11, 10) != complex<double>(0,0))
						ierr = MatSetValue(hamiltonian, rows-1, rows-14, blockHamiltonian[j][2].getData(11, 10), INSERT_VALUES);CHKERRQ(ierr);
				} //endif
			} //endfor
		} //endfor
	} //endfor

	for(int j=1; j<300; j++)
	{
		for(int k=0; k<12; k++)
		{
			for(int l=0; l<12; l++)
			{
				if(blockHamiltonian[j][0].getData(k, l) != complex<double>(0,0))
					ierr = MatSetValue(hamiltonian, k+12*(j-1), l+12*j, blockHamiltonian[j][0].getData(k, l), INSERT_VALUES);CHKERRQ(ierr);
			} //endfor
		} //endfor
	} //endfor

	for(int k=0; k<12; k++)
	{
		for (int l=0; l<12; l++)
		{
			if(blockHamiltonian[0][0].getData(k, l) != complex<double>(0,0))
				ierr = MatSetValue(hamiltonian, k, l+12*299, blockHamiltonian[0][0].getData(k, l), INSERT_VALUES);CHKERRQ(ierr);
			if(blockHamiltonian[299][2].getData(k, l) != complex<double>(0,0))
				ierr = MatSetValue(hamiltonian, k+12*299, l, blockHamiltonian[299][2].getData(k, l), INSERT_VALUES);CHKERRQ(ierr);
		} //endfor
	} //endfor

//******************************************************************************************************
	//Assemble PETSc Hamiltonian
	ierr = MatAssemblyBegin(hamiltonian, MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
	ierr = MatAssemblyEnd(hamiltonian, MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
	
	//Reset PETSc Solver
	ierr = KSPSetOperators(ksp, hamiltonian, hamiltonian, DIFFERENT_NONZERO_PATTERN);CHKERRQ(ierr);
	ierr = KSPSetUp(ksp);CHKERRQ(ierr);
	
	//Factor and solve hamiltonian * inverseHamiltonian = identityMat for inverseHamiltonian
	ierr = PCFactorGetMatrix(pc, &hamiltonian);CHKERRQ(ierr);
	ierr = MatMatSolve(hamiltonian, identityMat, inverseHamiltonian);CHKERRQ(ierr);
//******************************************************************************************************
	
	PetscFunctionReturn(0);
} //end InvertHamiltonian

#undef __FUNCT__
#define __FUNCT__ "RetrieveInformation"
PetscErrorCode RetrieveInformation (Mat inverseHamiltonian, double energy, double energyMin, double energyMax, double energyStep, int leadSize,
									double IVvalues[][2], ofstream& outFile, int isl[], int isr[], complex<double> gl[4][4], complex<double> gr[4][4])
{
	//PETSc Setup
	const PetscScalar *rowValues;
	PetscScalar trace;
	PetscInt nonZeros, numSubMatrices = 2, size;
	PetscErrorCode ierr;
	static Mat *localSubMatrices;
	static IS *indexSetRows, *indexSetColumns;
	ierr = MPI_Comm_size(PETSC_COMM_WORLD, &size);CHKERRQ(ierr);
	
	//Variable Declarations
	complex<double> gaml[4][4], gamrr[4][4];
	complex<double> prodl[4][4], prodr[4][4], zres[4][4];
	complex<double> fermi1, fermi2, fermif, transmissionFunction;
	complex<double> scaledTransmissionFunction;
	double densityOfStates = 0, therm = .03, vsdmax = .1, thop = -0.25;
	double emu1 = -vsdmax/2, emu2 = vsdmax/2;
	int currentStep = (energy < 0 ? (energy - energyMin) / energyStep + .5 : (energy - energyMin) / energyStep - .5);
	
	ierr = MatGetTrace(inverseHamiltonian, &trace);CHKERRQ(ierr);
	densityOfStates = -1 * trace.imag() / PI;
	
	if(size > 1) //parallel
	{
		//This only works if the rows and columns of gaml and gamrr are determined by arrays
		//Will not work for arbitrary gaml[i][j] = inverseHamiltonian[m][n]
		//if(energy == energyMin)
		//{
			ierr = PetscMalloc(numSubMatrices*sizeof(IS **), &indexSetRows);CHKERRQ(ierr);
			ierr = PetscMalloc(numSubMatrices*sizeof(IS **), &indexSetColumns);CHKERRQ(ierr);
			ierr = ISCreateGeneral(PETSC_COMM_SELF, leadSize, isl, indexSetRows);CHKERRQ(ierr);
			ierr = ISCreateGeneral(PETSC_COMM_SELF, leadSize, isr, indexSetRows+1);CHKERRQ(ierr);
			ierr = ISCreateGeneral(PETSC_COMM_SELF, leadSize, isr, indexSetColumns);CHKERRQ(ierr);
			ierr = ISCreateGeneral(PETSC_COMM_SELF, leadSize, isl, indexSetColumns+1);CHKERRQ(ierr);
			for(int i=0; i<numSubMatrices; i++)
			{
				ierr = ISSort(indexSetRows[i]);CHKERRQ(ierr);
				ierr = ISSort(indexSetColumns[i]);CHKERRQ(ierr);
			}
		//}
		//if(rank == 0) numSubMatrices = 2; else numSubMatrices = 0;
		ierr = MatGetSubMatrices(inverseHamiltonian, numSubMatrices, indexSetRows, indexSetColumns, MAT_INITIAL_MATRIX, &localSubMatrices);CHKERRQ(ierr);
		
		//Fill gaml and gamrr
		for(int i=0; i<leadSize; i++)
		{
			ierr = MatGetRow(localSubMatrices[0], i, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
			for(int j=0; j<leadSize; j++)
				gaml[i][j] = rowValues[j];
			ierr = MatRestoreRow(localSubMatrices[0], i, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
		} //endfor		
		for(int j=0; j<leadSize; j++)
		{
			ierr = MatGetRow(localSubMatrices[1], j, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
			for(int i=0; i<leadSize; i++)
				gamrr[i][j] = conj(rowValues[i]);
			ierr = MatRestoreRow(localSubMatrices[1], j, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
		} //endfor
		
		//Determine permutation needed
		int isrPerm[leadSize], islPerm[leadSize];
		for(int i=0; i<leadSize; i++)
		{
			isrPerm[i] = 0;
			islPerm[i] = 0;
			for(int j=0; j<leadSize; j++)
			{
				if(isr[j] < isr[i])	isrPerm[i]++;
				if(isl[j] < isl[i])	islPerm[i]++;
			}
		}
		
		//Permute rows (islPerm) and columns (isrPerm)
		for(int i=0; i<leadSize; i++)
		{
			int tempVar;
			if(islPerm[i] != i)
			{
				permuteRow(gaml, i, islPerm[i]);
				permuteRow(gamrr, i, islPerm[i]);
				tempVar = islPerm[i];
				islPerm[i] = i;
				islPerm[tempVar] = tempVar;
			}
			if(isrPerm[i] != i)
			{
				permuteColumn(gamrr, i, isrPerm[i]);
				permuteColumn(gaml, i, isrPerm[i]);
				tempVar = isrPerm[i];
				isrPerm[i] = i;
				isrPerm[tempVar] = tempVar;
			}
		}
		
		//Destroy objects when finished
		//if(energy == energyMax)
		//{
			for(int i=0; i<numSubMatrices; i++)
			{
				ierr = ISDestroy(indexSetRows[i]);CHKERRQ(ierr);
				ierr = ISDestroy(indexSetColumns[i]);CHKERRQ(ierr);
				ierr = MatDestroy(localSubMatrices[i]);CHKERRQ(ierr);
			}
			ierr = PetscFree(indexSetRows);CHKERRQ(ierr);
			ierr = PetscFree(indexSetColumns);CHKERRQ(ierr);
			ierr = PetscFree(localSubMatrices);CHKERRQ(ierr);
		//} //endif
	} //endif
	else //serial
	{
		for(int i=0; i<leadSize; i++)
		{
			ierr = MatGetRow(inverseHamiltonian, isl[i]-1, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
			for(int j=0; j<leadSize; j++)
				gaml[i][j] = rowValues[isr[j]-1];
			ierr = MatRestoreRow(inverseHamiltonian, isl[i]-1, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
		} //endfor
		
		for(int j=0; j<leadSize; j++)
		{
			ierr = MatGetRow(inverseHamiltonian, isr[j]-1, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
			for(int i=0; i<leadSize; i++)
				gamrr[i][j] = conj(rowValues[isl[i]-1]);
			ierr = MatRestoreRow(inverseHamiltonian, isr[j]-1, &nonZeros, PETSC_NULL, &rowValues);CHKERRQ(ierr);
		} //endfor
	} //endelse
	
	//Multiply the previous four leadSize x leadSize matrices together to get the transmissionFunction
	for(int i=0; i<leadSize; i++)
	{
		for (int j=0; j<leadSize; j++)
		{
			for(int k=0; k<leadSize; k++)
			{
				prodl[i][j] += -gl[i][k].imag() * gaml[k][j];
				prodr[i][j] += -gr[i][k].imag() * gamrr[k][j];
			} //endfor
		} //endfor
	} //endfor
	
	for(int i=0; i<leadSize; i++)
	{
		for(int j=0; j<leadSize; j++)
		{
			zres[i][j] = complex<double>(0,0);
			for(int k=0; k<leadSize; k++)
			{
				zres[i][j] += prodl[i][k] * prodr[k][j];
			} //endfor
		} //endfor
	} //endfor
	
	for(int i=0; i<leadSize; i++)
		transmissionFunction += zres[i][i];
	scaledTransmissionFunction = 4*pow(thop,4)*transmissionFunction;
	
	fermi1 = 1.0 / (exp((energy-emu1)/therm)+1.0);
	fermi2 = 1.0 / (exp((energy-emu2)/therm)+1.0);
	fermif = 2 * abs(transmissionFunction) * (fermi1-fermi2);
	
	//Output to terminal
	ierr = PetscPrintf(PETSC_COMM_WORLD, "%.5f\t%.5e\t   (%.5e, %.5e)\t%.5e", energy, densityOfStates, scaledTransmissionFunction.real(), scaledTransmissionFunction.imag(), fermif.real());CHKERRQ(ierr);
	
	//Output to file
	outFile.precision(5);
	outFile << fixed << energy << " \t" << scientific << densityOfStates << "\t" << scaledTransmissionFunction << "\t" << fermif.real() << endl;
	outFile.unsetf(ios_base::floatfield);
	
	//Fill array for later integration
	IVvalues[currentStep][0] = energy;
	IVvalues[currentStep][1] = 4 * pow(thop,4) * abs(transmissionFunction);	

	PetscFunctionReturn(0);
} //end RetrieveInformation

void permuteRow (complex<double> array[4][4], int rowA, int rowB)
{
	complex<double> tempVar;
	for(int j=0; j<4; j++)
	{
		tempVar = array[rowA][j];
		array[rowA][j] = array[rowB][j];
		array[rowB][j] = tempVar;
	}
} //end permuteRow

void permuteColumn (complex<double> array[4][4], int colA, int colB)
{
	complex<double> tempVar;
	for(int i=0; i<4; i++)
	{
		tempVar = array[i][colA];
		array[i][colA] = array[i][colB];
		array[i][colB] = tempVar;
	}
} //end permuteColumn

#endif /* INVERT_H_ */
